import os
import sys

from datetime import datetime


def cmd_str():
    """ get cmd that starts current script"""
    args = ' '.join(sys.argv)
    return f'python {args}'


def datetime_str():
    dtstr = datetime.now().strftime("%Y-%m-%d-%H:%M:%S")
    return dtstr


def instantiate(path, *args, **kwargs):
    """ instantiate a class or call a function with given args and kwargs 
        from `anywhere`.

        Example:
            `instantiate('module.a.B', t=1e-3)`
            return `module.a.B(t=1e-3)`
    """
    from pydoc import locate
    obj = locate(path)
    return obj(*args, **kwargs)


def auto_rename(path, ignore_ext=False):
    count = 1
    new_path = path
    while True:
        if not os.path.exists(new_path):
            return new_path
        file_name = os.path.basename(path)
        try:
            name, ext = file_name.split('.')
            new_name = f'{name}_{count}.{ext}'
        except:
            new_name = f'{file_name}_{count}'
        if ignore_ext:
            new_name = f'{file_name}_{count}'
        new_path = os.path.join(os.path.dirname(path), new_name)
        count += 1



import cv2
from typing import List

import numpy as np
import matplotlib.pyplot as plt


def add_text(image, text):
    image = cv2.resize(image, (512, 512))
    # overlay space
    x, y, w, h = 0, 0, 250, 50

    # alpha, the 4th channel of the image
    alpha = 0.5

    overlay = image.copy()
    output = image.copy()

    # corner
    cv2.rectangle(overlay, (x, x), (x + w, y + h), (0, 0, 0), -1)

    # putText
    cv2.putText(overlay, text, (x + 50, y + int(h - 20)), cv2.FONT_HERSHEY_SIMPLEX, 1, (255, 255, 255), 3)

    # apply the overlay
    cv2.addWeighted(overlay, alpha, output, 1 - alpha, 0, output)
    return output


def imshow(*imgs: List[np.ndarray],
           maxcol: int = 3,
           gray: bool = False,
           titles: List[str] = None,
           off_axis: bool = True) -> None:
    """
    Display one or more images in a grid with customizable parameters such as 
    maximum number of columns, grayscale, and titles.

    Args:
      imgs (List[np.ndarray]): a list of images.
      maxcol (int): The maximum number of columns to display the images in. If there are more images than
        maxcol, they will be displayed in multiple rows. The default value is 3, defaults to 3 (optional)
      gray (bool): A boolean parameter that determines whether the image(s) should be displayed in
        grayscale or in color. If set to True, the images will be displayed in grayscale. If set to False,
        the images will be displayed in color, defaults to False (optional)
      titles (List[str]): titles is a list of strings that contains the titles for each image being displayed. If titles is None, then no titles will be displayed
      off_axis (bool): whether to remove axis in the images.
    """
    if len(imgs) != 1:
        plt.figure(figsize=(10, 5), dpi=300)
    row = (len(imgs) - 1) // maxcol + 1
    col = maxcol if len(imgs) >= maxcol else len(imgs)
    for idx, img in enumerate(imgs):
        if img.max() > 2: img = img / 255
        img = img.clip(0, 1)
        if gray: plt.gray()
        plt.subplot(row, col, idx + 1)
        plt.imshow(img)
        if titles is not None: plt.title(titles[idx])
        if off_axis: plt.axis('off')
    plt.show()



import cv2
import torch
import numpy as np
from PIL import Image


def rgb2bgr(img):
    img = cv2.cvtColor(img, cv2.COLOR_RGB2BGR)
    return img


def gray2rgb(img):
    img = cv2.cvtColor(img, cv2.COLOR_GRAY2RGB)
    return img


def bgr2rgb():
    img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)
    return img


def to_pillow_image(image: np.ndarray):
    return Image.fromarray(image)


def batchify(out):
    """
    Take an input tensor and return it as a batch tensor with an additional dimension.

    Args:
      out: The input tensor that needs to be batchified.

    Returns:
      a PyTorch tensor. If the input tensor `out` is already a delta-prox tensor, it is returned as is.
    Otherwise, it returns a tensor with BCHW format if the input tensor has 3 dimensions with HWC
    format.
    """
    if len(out.shape) == 3 and (out.shape[2] == 1 or out.shape[2] == 3):
        out = out.permute(2, 0, 1)
    out = out.unsqueeze(0)
    return out


def to_torch_tensor(x, batch=False):
    """
    Convert a given input to a torch tensor and add a batch dimension if specified.

    Args:
      x: The input data that needs to be converted to a torch tensor.
      batch: A boolean parameter that indicates whether to add a batch dimension to the tensor or not.
    If set to True, a batch dimension will be added to the tensor. If set to False, no batch dimension
    will be added. Defaults to False

    Returns:
      a PyTorch tensor. If the input is already a dp.tensor, it returns the input as is. If the input is a
    numpy array, it converts it to a dp.tensor. If the input is neither a torch tensor nor a numpy
    array, it assumes it is a dp.tensor and returns it. If the batch parameter is True, it adds a batch
    dimension to the tensor and returns
    """
    if isinstance(x, torch.Tensor):
        out = x
    elif isinstance(x, np.ndarray):
        out = dp.tensor(x.copy())
    else:
        out = dp.tensor(x)

    if batch:
        if len(out.shape) == 3 and (out.shape[2] == 1 or out.shape[2] == 3):
            out = out.permute(2, 0, 1)
        if len(out.shape) < 4:
            out = out.unsqueeze(0)

    out.is_dp_tensor = True
    return out


def debatchify(out, squeeze):
    """
    Debatchify a tensor by squeezing and/or transposing its dimensions,
    supporting multiple tensor format transforms (BCHW -> CHW | CHW -> HWC | HWC -> HW)

    Args:
      out: The output tensor that needs to be transformed. It could be in the format of BCHW, CHW, or
    HWC.
      squeeze: A boolean parameter that determines whether to convert tensor format with C = 1 to HW
    format. If set to True, it will remove the singleton dimension and return a tensor with shape (H, W)
    instead of (H, W, 1). 

    Returns:
      the tensor `out` with simplified format.
    """
    if len(out.shape) == 4:
        out = out.squeeze(0)  # BCHW -> CHW
    if len(out.shape) == 3:
        if out.shape[0] == 3 or out.shape[0] == 1:
            out = out.transpose(1, 2, 0)  # CHW -> HWC
        if out.shape[2] == 1 and squeeze:
            out = out.squeeze(2)  # HWC -> HW
    return out


def to_ndarray(x, debatch=False, squeeze=False):
    """
    Convert a given input into a numpy array and optionally remove any batch dimensions.

    Args:
      x: The input data that needs to be converted to a numpy array.
      debatch: A boolean parameter that specifies whether to remove the batch dimension from the input
    tensor or not. If set to True, the function will call the `debatchify` function to remove the batch
    dimension. If set to False, the function will return the input tensor as is. Defaults to False
      squeeze: The `squeeze` parameter is a boolean parameter that determines whether to convert a
    tensor format with C = 1 to HW format. 

    Returns:
      a numpy array. If `debatch` is True, the output is passed through the `debatchify` function with
    `squeeze` before being returned.
    """
    if isinstance(x, torch.Tensor):
        out = x.detach().cpu().numpy()
    elif isinstance(x, np.ndarray):
        out = x.astype('float32')
    else:
        out = np.array(x)
    if debatch:
        out = debatchify(out, squeeze)
    return out



import os
import glob
import tqdm
import cv2
import json
import varname
from PIL import Image

import torch
import numpy as np

from rich.console import Console

# inspect array like object x and report stats


def lo(*xs, verbose=0):

    console = Console()

    def _lo(x, name):

        if isinstance(x, np.ndarray):
            # general stats
            text = ""
            text += f"[orange1]Array {name}[/orange1] {x.shape} {x.dtype}"
            if x.size > 0:
                text += f" ∈ [{x.min()}, {x.max()}]"
            if verbose >= 1:
                text += f" μ = {x.mean()} σ = {x.std()}"
            # detect abnormal values
            if np.isnan(x).any():
                text += "[red] NaN![/red]"
            if np.isinf(x).any():
                text += "[red] Inf![/red]"
            console.print(text)

            # show values if shape is small or verbose is high
            if x.size < 50 or verbose >= 2:
                # np.set_printoptions(precision=4)
                print(x)

        elif torch.is_tensor(x):
            # general stats
            text = ""
            text += f"[orange1]Tensor {name}[/orange1] {x.shape} {x.dtype} {x.device}"
            if x.numel() > 0:
                text += f"∈ [{x.min().item()}, {x.max().item()}]"
            if verbose >= 1:
                text += f" μ = {x.mean().item()} σ = {x.std().item()}"
            # detect abnormal values
            if torch.isnan(x).any():
                text += "[red] NaN![/red]"
            if torch.isinf(x).any():
                text += "[red] Inf![/red]"
            console.print(text)

            # show values if shape is small or verbose is high
            if x.numel() < 50 or verbose >= 2:
                # np.set_printoptions(precision=4)
                print(x)

        else:  # other type, just print them
            console.print(f"[orange1]{type(x)} {name}[/orange1] {x}")

    # inspect names
    for i, x in enumerate(xs):
        _lo(x, varname.argname(f"xs[{i}]"))


def read_json(path):
    with open(path, "r") as f:
        return json.load(f)


def write_json(path, x):
    with open(path, "w") as f:
        json.dump(x, f, indent=2)


def read_image(path, mode="float", order="RGB"):

    if mode == "pil":
        return Image.open(path)

    img = cv2.imread(path, cv2.IMREAD_UNCHANGED)

    # cvtColor
    if order == "RGB":
        if len(img.shape) == 3:
            if img.shape[2] == 4:
                img = cv2.cvtColor(img, cv2.COLOR_BGRA2RGBA)
            else:
                img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)

    # mode
    if "float" in mode:
        return img.astype(np.float32) / 255
    elif "tensor" in mode:
        return torch.from_numpy(img.astype(np.float32) / 255)
    else:  # uint8
        return img


def write_image(path, img, order="RGB"):

    if torch.is_tensor(img):
        img = img.detach().cpu().numpy()

    if img.dtype == np.float32 or img.dtype == np.float64:
        img = (img * 255).astype(np.uint8)

    # cvtColor
    if order == "RGB":
        if len(img.shape) == 3:
            if img.shape[2] == 4:
                img = cv2.cvtColor(img, cv2.COLOR_RGBA2BGRA)
            else:
                img = cv2.cvtColor(img, cv2.COLOR_RGB2BGR)

    dir_path = os.path.dirname(path)
    if dir_path != '' and not os.path.exists(dir_path):
        os.makedirs(os.path.dirname(path), exist_ok=True)
    cv2.imwrite(path, img)


def load_file_from_url(url, model_dir=None, progress=True, file_name=None):

    from torch.hub import download_url_to_file, get_dir
    from urllib.parse import urlparse

    """Load file form http url, will download models if necessary.

    Ref:https://github.com/1adrianb/face-alignment/blob/master/face_alignment/utils.py

    Args:
        url (str): URL to be downloaded.
        model_dir (str): The path to save the downloaded model. Should be a full path. If None, use pytorch hub_dir.
            Default: None.
        progress (bool): Whether to show the download progress. Default: True.
        file_name (str): The downloaded file name. If None, use the file name in the url. Default: None.

    Returns:
        str: The path to the downloaded file.
    """
    if model_dir is None:  # use the pytorch hub_dir
        hub_dir = get_dir()
        model_dir = os.path.join(hub_dir, "checkpoints")

    os.makedirs(model_dir, exist_ok=True)

    parts = urlparse(url)
    filename = os.path.basename(parts.path)
    if file_name is not None:
        filename = file_name
    cached_file = os.path.abspath(os.path.join(model_dir, filename))
    if not os.path.exists(cached_file):
        print(f'Downloading: "{url}" to {cached_file}\n')
        download_url_to_file(url, cached_file, hash_prefix=None, progress=progress)
    return cached_file


def is_format(f, format):
    return os.path.splitext(f)[1].lower() in format


def batch_process_files(
    process_fn, path, out_path,
    overwrite=False,
    in_format=[".jpg", ".jpeg", ".png"],
    out_format=None,
    color_order="RGB",
    **kwargs
):

    if os.path.isdir(path):
        file_paths = glob.glob(os.path.join(path, "*"))
        file_paths = [f for f in file_paths if is_format(f, in_format)]
    else:
        file_paths = [path]

    if os.path.dirname(out_path) != '':
        os.makedirs(os.path.dirname(out_path), exist_ok=True)

    for file_path in tqdm.tqdm(file_paths):
        try:

            if len(file_paths) == 1:
                file_out_path = out_path
            else:
                file_out_path = os.path.join(out_path, os.path.basename(file_path))

            if out_format is not None:
                file_out_path = os.path.splitext(file_out_path)[0] + out_format

            if os.path.exists(file_out_path) and not overwrite:
                print(f"[INFO] ignoring {file_path} --> {file_out_path}")
                continue

            # dispatch suitable loader and writer
            # only support image and text file
            if is_format(file_path, ['.jpg', '.jpeg', '.png']):
                input = read_image(file_path, mode="uint8", order=color_order)
            else:
                with open(file_path, "r") as f:
                    input = f.read()

            output = process_fn(input, **kwargs)

            # only support image, npy or text file
            if is_format(file_out_path, ['.jpg', '.jpeg', '.png']):
                write_image(file_out_path, output, order=color_order)
            elif is_format(file_out_path, ['.npy']):
                np.save(file_out_path, output)
            else:
                with open(file_out_path, "w") as f:
                    f.write(output)

        except Exception as e:
            print(f"[Error] when processing {file_path} --> {file_out_path}")
            print(e)



from setuptools import setup, find_packages

setup(
    name='dola',
    author='Zeqiang Lai',
    author_email='laizeqiang@outlook.com',
    packages=find_packages(),
    version='0.0.4',
    include_package_data=True,
)



import urllib.request
from tqdm import tqdm


class DownloadProgressBar(tqdm):
    def update_to(self, b=1, bsize=1, tsize=None):
        if tsize is not None:
            self.total = tsize
        self.update(b * bsize - self.n)


def download_url(url: str, output_path: str) -> None:
    with DownloadProgressBar(
        unit="B", unit_scale=True, miniters=1, desc=url.split("/")[-1]
    ) as t:
        urllib.request.urlretrieve(
            url, filename=output_path, reporthook=t.update_to
        )


from typing import Optional

import cv2
import numpy as np

MODES = {
    'cubic': cv2.INTER_CUBIC,
    'nearest': cv2.INTER_NEAREST,
}


def imresize(data, size, mode='cubic'):
    """ data ndarray (H,W,C) or (H,W)
        size (height, width)
    """
    height, width = size
    mode = MODES[mode]
    out = cv2.resize(data, dsize=(width, height), interpolation=mode)
    if len(data.shape) == 3 and data.shape[-1] == 1:
        out = out[:,:,None]
    return out


def ceil_modulo(x, mod):
    if x % mod == 0:
        return x
    return (x // mod + 1) * mod


def pad_img_to_modulo(
    img: np.ndarray,
    mod: int,
    square: bool = False,
    min_size: Optional[int] = None,
):
    """

    Args:
        img: [H, W, C]
        mod:
        square: 是否为正方形
        min_size:

    Returns:

    """
    if len(img.shape) == 2:
        img = img[:, :, np.newaxis]
    height, width = img.shape[:2]
    out_height = ceil_modulo(height, mod)
    out_width = ceil_modulo(width, mod)

    if min_size is not None:
        assert min_size % mod == 0
        out_width = max(min_size, out_width)
        out_height = max(min_size, out_height)

    if square:
        max_size = max(out_height, out_width)
        out_height = max_size
        out_width = max_size

    return np.pad(
        img,
        ((0, out_height - height), (0, out_width - width), (0, 0)),
        mode="symmetric",
    )



from skimage.metrics import peak_signal_noise_ratio, structural_similarity
import functools
import numpy as np

FORMAT_HWC = 'hwc'
FORMAT_CHW = 'chw'
DATA_FORMAT = FORMAT_HWC

__all__ = [
    'psnr',
    'ssim',
    'sam',
    'mpsnr',
    'mssim'
]

def set_data_format(format):
    if format.lower() != FORMAT_HWC and format.lower() != FORMAT_CHW:
        raise ValueError('Invalid data format, choose from '
                         'torchlight.metrics.HWC or torchlight.metrics.CHW')
    global DATA_FORMAT
    DATA_FORMAT = format


def CHW2HWC(func):
    @functools.wraps(func)
    def warpped(output, target, *args, **kwargs):
        if DATA_FORMAT == FORMAT_CHW:
            output = output.transpose(1, 2, 0)
            target = target.transpose(1, 2, 0)
        return func(output, target, *args, **kwargs)
    return warpped


def torch2numpy(func):
    @functools.wraps(func)
    def warpped(output, target, *args, **kwargs):
        if not isinstance(output, np.ndarray):
            output = output.detach().cpu().numpy()
            target = target.detach().cpu().numpy()
        return func(output, target, *args, **kwargs)
    return warpped


def bandwise(func):
    @functools.wraps(func)
    def warpped(output, target, *args, **kwargs):
        if DATA_FORMAT == FORMAT_CHW:
            C = output.shape[-3]
            total = 0
            for ch in range(C):
                x = output[ch, :, :]
                y = target[ch, :, :]
                total += func(x, y, *args, **kwargs)
            return total / C
        else:
            C = output.shape[-1]
            total = 0
            for ch in range(C):
                x = output[:, :, ch]
                y = target[:, :, ch]
                total += func(x, y, *args, **kwargs)
            return total / C
    return warpped


def enable_batch_input(reduce=True):
    def inner(func):
        @functools.wraps(func)
        def warpped(output, target, *args, **kwargs):
            if len(output.shape) == 4:
                b = output.shape[0]
                out = [func(output[i], target[i]) for i in range(b)]
                if reduce:
                    return sum(out) / len(out)
                return out
            return func(output, target, *args, **kwargs)
        return warpped
    return inner


# raw psnr, ssim, sam assume HWC format


@torch2numpy
@enable_batch_input()
@CHW2HWC
def psnr(output, target, data_range=1):
    return peak_signal_noise_ratio(target, output, data_range=data_range)


@torch2numpy
@enable_batch_input()
@CHW2HWC
def ssim(img1, img2, **kwargs):
    return structural_similarity(img1, img2, channel_axis=2, **kwargs)


@torch2numpy
@enable_batch_input()
@CHW2HWC
def sam(img1, img2, eps=1e-8):
    """
    Spectral Angle Mapper which defines the spectral similarity between two spectra
    """
    tmp1 = np.sum(img1 * img2, axis=2) + eps
    tmp2 = np.sqrt(np.sum(img1**2, axis=2)) + eps
    tmp3 = np.sqrt(np.sum(img2**2, axis=2)) + eps
    tmp4 = tmp1 / tmp2 / tmp3
    angle = np.arccos(tmp4.clip(-1, 1))
    return np.mean(np.real(angle))


@torch2numpy
@enable_batch_input()
@bandwise
def mpsnr(output, target, data_range=1):
    return peak_signal_noise_ratio(target, output, data_range=data_range)


@torch2numpy
@enable_batch_input()
@bandwise
def mssim(img1, img2, **kwargs):
    return structural_similarity(img1, img2, **kwargs)
